/**********************************************************
Camera class created by Jeremy Paskali

camera.cpp
**********************************************************/
#include"camera.h"

CCamera::CCamera(void)
{
	eye.set(0,0,0);
	u.set(0,0,0);
	v.set(0,0,0);
	n.set(0,0,0);
	viewAngle = aspect = 0;
	nearDist = farDist = 0;
}

// forwards if direction = 1
// backwards if direction = -1
void CCamera::move(int direction)
{
	eye.x = eye.x - direction*(n.x)*0.1;
	eye.y = eye.y - direction*(n.y)*0.1;
	eye.z = eye.z - direction*(n.z)*0.1;
	setModelViewMatrix();
}

void CCamera::pitch(float angle)
{
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	Vector3 t = v;  //remember old v
	v.set(cs*t.x - sn*n.x, cs*t.y - sn*n.y, cs*t.z - sn*n.z);
	n.set(sn*t.x + cs*n.x, sn*t.y + cs*n.y, sn*t.z + cs*n.z);
	setModelViewMatrix();
}

void CCamera::yaw(float angle)
{
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	Vector3 t = u;  //remember old u
	u.set(cs*t.x - sn*n.x, cs*t.y - sn*n.y, cs*t.z - sn*n.z);
	n.set(sn*t.x + cs*n.x, sn*t.y + cs*n.y, sn*t.z + cs*n.z);
	setModelViewMatrix();
}

void CCamera::roll(float angle)
{
	float cs = cos(3.14159265/180 * angle);
	float sn = sin(3.14159265/180 * angle);
	Vector3 t = u;  //remember old u
	u.set(cs*t.x - sn*v.x, cs*t.y - sn*v.y, cs*t.z - sn*v.z);
	v.set(sn*t.x + cs*v.x, sn*t.y + cs*v.y, sn*t.z + cs*v.z);
	setModelViewMatrix();
}

void CCamera::slide(int direction)
{
	eye.x = eye.x - direction*(u.x)*0.1;
	eye.y = eye.y - direction*(u.y)*0.1;
	eye.z = eye.z - direction*(u.z)*0.1;
	setModelViewMatrix();
}

void CCamera::rise(int direction)
{
	eye.x = eye.x + direction*(v.x)*0.1;
	eye.y = eye.y + direction*(v.y)*0.1;
	eye.z = eye.z + direction*(v.z)*0.1;
	setModelViewMatrix();
}

void CCamera::setShape(float vAng, float asp, float nearD, float farD)
{
	viewAngle = vAng;
	aspect = asp;
	nearDist = nearD;
	farDist = farD;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( viewAngle, aspect, nearDist, farDist);
	//glOrtho(-1, 1, -1, 1, 12, -1);
}

// EPN: Added because setShape does not call setModelView Matrix.
// I didn't want to expose setModelView, so I created another method
// (this one) to do what I needed
void CCamera::modifyShape(float vAng, float asp, float nearD, float farD)
{
	viewAngle = vAng;
	aspect = asp;
	nearDist = nearD;
	farDist = farD;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( viewAngle, aspect, nearDist, farDist);
	//glOrtho(-2, 2, -2, 2, 12, -2);
	setModelViewMatrix();
}


void CCamera::setModelViewMatrix(void)
{
	//load modelview matrix with existing camera values
	float m[16];
	Vector3 eVec(eye.x, eye.y, eye.z);  // a vector version of the eye
	m[0] = u.x; m[4] = u.y; m[8]  = u.z; m[12] = -eVec.dot(u);
	m[1] = v.x; m[5] = v.y; m[9]  = v.z; m[13] = -eVec.dot(v);
	m[2] = n.x; m[6] = n.y; m[10] = n.z; m[14] = -eVec.dot(n);
	m[3] = 0;   m[7] = 0;   m[11] = 0;   m[15] = 1.0;
	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(m);  // Load OpenGL's model view matrix
}

void CCamera::set(Point3 Eye, Point3 look, Vector3 up)
{
	// create a modelview matrix and send it to OPENGL
	eye.set(Eye); // store the given eye position
	n.set(eye.x - look.x, eye.y - look.y, eye.z - look.z); // make n
	u.set(up.cross(n)); // make u = up X n (up cross product n)
	n.normalize(); u.normalize(); //make them unit length
	v.set(n.cross(u));  //make v = n X u (n cross product u)
	setModelViewMatrix();  //tell OpenGL
}

/***************************************************************
Vector3 class provided by 
Computer Graphics with OpenGL, by FS Hill JR
and Prof. Ed Nelson
Modified and Enhanced by Jeremy Paskali
***************************************************************/

void Vector3::normalize(void)
{
	float normalize = sqrt(x*x + y*y + z*z);
	x = x / normalize;
	y = y / normalize;
	z = z / normalize;
}

// return this vector cross vector b
Vector3 Vector3::cross(Vector3 b)
{
	// a x b = (ay*bz - az*by)i + (az*bx - ax*bz)j + (ax*by - ay*bx)k
	Vector3 result;
	result.x = y*b.z - z*b.y;
	result.y = z*b.x - x*b.z;
	result.z = x*b.y - y*b.x;

	return result;
}

// return this vector dot vector b
float Vector3::dot(Vector3 b)
{
	return (x*b.x + y*b.y + z*b.z);
}

void Mesh::makeSphere( float radius )
{
	// x = r * cos( phi ) * cos( theta )
	// y = r * cos( phi ) * sin( theta )
	// z = r * sin( phi )

	int z = 0;
	float PI = 3.14159265;
	float phi = -PI / 2.0, theta = 0;

	// initial point (south pole)
	pt[z].set( radius * cos( phi ) * cos( theta ),
	           radius * cos( phi ) * sin( theta ),
			   radius * sin( phi ) );
	norm[z].set(pt[z].x, pt[z].y, pt[z].z);
	norm[z].normalize();

	phi += PI / 6.0;

	for( int p = 0; p < 6; p++ )
	{
		for( int u = 0; u < 12; u++ )
		{
			z++;
			theta += PI / 6.0;
			pt[z].set( radius * cos( phi ) * cos( theta ),
					   radius * cos( phi ) * sin( theta ),
					   radius * sin( phi ) );
			norm[z].set(pt[z].x, pt[z].y, pt[z].z);
			norm[z].normalize();
		}
		theta = 0;
		phi += PI / 6.0;
	}
	z++;
	

	

	// final point (north pole)
	pt[z].set( radius * cos( phi ) * cos( theta ),
			   radius * cos( phi ) * sin( theta ),
			   radius * sin( phi ) );
	norm[z].set(pt[z].x, pt[z].y, pt[z].z);
	norm[z].normalize();

	// The point array is full of points we need to
	// normals are set
	// create faces with vertices

	// Bottom ring of triangles
	int i = 0, j = 0, k = 0;
	for( i = 1; i < 12; i++ )
	{
		// triangles and their vertices
		face[i].nVerts = 4;

		face[i].vert[0].vertIndex = 0;
		face[i].vert[0].normIndex = 0;

		face[i].vert[1].vertIndex = i+1;
		face[i].vert[1].normIndex = i+1;

		face[i].vert[2].vertIndex = i;
		face[i].vert[2].normIndex = i;

		face[i].vert[3].vertIndex = 0;
		face[i].vert[3].normIndex = 0;
	}

	face[i].nVerts = 4;

	face[i].vert[0].vertIndex = 0;
	face[i].vert[0].normIndex = 0;

	face[i].vert[1].vertIndex = 1;
	face[i].vert[1].normIndex = 1;

	face[i].vert[2].vertIndex = 12;
	face[i].vert[2].normIndex = 12;

	face[i].vert[3].vertIndex = 0;
	face[i].vert[3].normIndex = 0;

	i++;

	// twelve quadrilaterals across each row

	// First row of quads
	for( j = 1; j < 12; j++ )
	{
		// Vertices and normals
		face[i].nVerts = 5;

		face[i].vert[0].vertIndex = j+1;
		face[i].vert[0].normIndex = j+1;

		face[i].vert[1].vertIndex = j+13;
		face[i].vert[1].normIndex = j+13;

		face[i].vert[2].vertIndex = j+12;
		face[i].vert[2].normIndex = j+12;

		face[i].vert[3].vertIndex = j;
		face[i].vert[3].normIndex = j;

		face[i].vert[4].vertIndex = j+1;
		face[i].vert[4].normIndex = j+1;

		i++;
	}

	face[i].nVerts = 5;

	face[i].vert[0].vertIndex = 1;
	face[i].vert[0].normIndex = 1;

	face[i].vert[1].vertIndex = 13;
	face[i].vert[1].normIndex = 13;

	face[i].vert[2].vertIndex = 24;
	face[i].vert[2].normIndex = 24;

	face[i].vert[3].vertIndex = 12;
	face[i].vert[3].normIndex = 12;

	face[i].vert[4].vertIndex = 1;
	face[i].vert[4].normIndex = 1;

	i++;

	// Second row of quads
	for( j = 13; j < 24; j++ )
	{
		// Vertices and normals
		face[i].nVerts = 5;

		face[i].vert[0].vertIndex = j+1;
		face[i].vert[0].normIndex = j+1;

		face[i].vert[1].vertIndex = j+13;
		face[i].vert[1].normIndex = j+13;

		face[i].vert[2].vertIndex = j+12;
		face[i].vert[2].normIndex = j+12;

		face[i].vert[3].vertIndex = j;
		face[i].vert[3].normIndex = j;

		face[i].vert[4].vertIndex = j+1;
		face[i].vert[4].normIndex = j+1;

		i++;
	}

	face[i].nVerts = 5;

	face[i].vert[0].vertIndex = 13;
	face[i].vert[0].normIndex = 13;

	face[i].vert[1].vertIndex = 25;
	face[i].vert[1].normIndex = 25;

	face[i].vert[2].vertIndex = 36;
	face[i].vert[2].normIndex = 36;

	face[i].vert[3].vertIndex = 24;
	face[i].vert[3].normIndex = 24;

	face[i].vert[4].vertIndex = 13;
	face[i].vert[4].normIndex = 13;

	i++;

	// Third row of quads
	for( j = 25; j < 36; j++ )
	{
		// Vertices and normals
		face[i].nVerts = 5;

		face[i].vert[0].vertIndex = j+1;
		face[i].vert[0].normIndex = j+1;

		face[i].vert[1].vertIndex = j+13;
		face[i].vert[1].normIndex = j+13;

		face[i].vert[2].vertIndex = j+12;
		face[i].vert[2].normIndex = j+12;

		face[i].vert[3].vertIndex = j;
		face[i].vert[3].normIndex = j;

		face[i].vert[4].vertIndex = j+1;
		face[i].vert[4].normIndex = j+1;

		i++;
	}

	face[i].nVerts = 5;

	face[i].vert[0].vertIndex = 25;
	face[i].vert[0].normIndex = 25;

	face[i].vert[1].vertIndex = 37;
	face[i].vert[1].normIndex = 37;

	face[i].vert[2].vertIndex = 48;
	face[i].vert[2].normIndex = 48;

	face[i].vert[3].vertIndex = 36;
	face[i].vert[3].normIndex = 36;

	face[i].vert[4].vertIndex = 25;
	face[i].vert[4].normIndex = 25;

	i++;

	// Fourth row of quads
	for( j = 37; j < 48; j++ )
	{
		// Vertices and normals
		face[i].nVerts = 5;

		face[i].vert[0].vertIndex = j+1;
		face[i].vert[0].normIndex = j+1;

		face[i].vert[1].vertIndex = j+13;
		face[i].vert[1].normIndex = j+13;

		face[i].vert[2].vertIndex = j+12;
		face[i].vert[2].normIndex = j+12;

		face[i].vert[3].vertIndex = j;
		face[i].vert[3].normIndex = j;

		face[i].vert[4].vertIndex = j+1;
		face[i].vert[4].normIndex = j+1;

		i++;
	}

	face[i].nVerts = 4;

	face[i].vert[0].vertIndex = 37;
	face[i].vert[0].normIndex = 37;

	face[i].vert[1].vertIndex = 49;
	face[i].vert[1].normIndex = 49;

	face[i].vert[2].vertIndex = 60;
	face[i].vert[2].normIndex = 60;

	face[i].vert[3].vertIndex = 48;
	face[i].vert[3].normIndex = 48;

	face[i].vert[4].vertIndex = 37;
	face[i].vert[4].normIndex = 37;

	i++;

	// Top triangles
	for( j = 49; j < 60; j++ )
	{
		// Vertices and normals
		face[i].nVerts = 4;

		face[i].vert[1].vertIndex = 61;
		face[i].vert[1].normIndex = 61;

		face[i].vert[2].vertIndex = j;
		face[i].vert[2].normIndex = j;

		face[i].vert[3].vertIndex = j+1;
		face[i].vert[3].normIndex = j+1;

		face[i].vert[0].vertIndex = 61;
		face[i].vert[0].normIndex = 61;

		i++;
	}
	
	face[i].nVerts = 4;

	face[i].vert[0].vertIndex = 61;
	face[i].vert[0].normIndex = 61;

	face[i].vert[1].vertIndex = 60;
	face[i].vert[1].normIndex = 60;

	face[i].vert[2].vertIndex = 49;
	face[i].vert[2].normIndex = 49;

	face[i].vert[3].vertIndex = 61;
	face[i].vert[3].normIndex = 61;

	i++;

	numFaces = i;
}

void Mesh::makePyramid(float percent)
{
	int g = 0;
	int f = 0;

	float x, by, z, ty;

	for(int c = 0; c < 100; c += 2)
	{
		// bottom shelf coords
		x = percent*float(100-c)/2.0;
		by = percent*(c - 1);
		z = percent*float(100-c)/2.0;

		// top shelf coords
		ty = percent*(c + 1);

		// Each layer of z is a rectangular polygon
		// This means there will be six faces
		// and eight points

		// Locations given in respect to 2D surface looking
		// down negative y-axis

		/**************************/
		/****** BOTTOM SHELF ******/
		/**************************/

		face[f].nVerts = 5;
		// 1st point
		// Top Right on Bottom shelf
		pt[g].set(x, by, z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[3].normIndex = g;
		face[f].vert[3].vertIndex = g;
		g++;

		// 2nd Point
		// Top Left on Bottom shelf
		pt[g].set(-x, by, z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[2].normIndex = g;
		face[f].vert[2].vertIndex = g;
		g++;

		// 3rd Point
		// lower Left on bottom shelf
		pt[g].set(-x, by, -z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[1].normIndex = g;
		face[f].vert[1].vertIndex = g;
		g++;

		// 4th point
		// lower right on bottom shelf
		pt[g].set(x, by, -z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[0].normIndex = g;
		face[f].vert[0].vertIndex = g;
		face[f].vert[4].normIndex = g;
		face[f].vert[4].vertIndex = g;
		g++;

		f++;
		
		/**************************/
		/****** UPPER SHELF *******/
		/**************************/

		face[f].nVerts = 5;
		// 5th point
		// top right on upper shelf
		pt[g].set(x, ty, z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[3].normIndex = g;
		face[f].vert[3].vertIndex = g;
		
		g++;

		// 6th Point
		// Top Left on upper shelf
		pt[g].set(-x, ty, z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[2].normIndex = g;
		face[f].vert[2].vertIndex = g;
		g++;

		// 7th Point
		// lower Left on upper shelf
		pt[g].set(-x, ty, -z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[1].normIndex = g;
		face[f].vert[1].vertIndex = g;
		g++;

		// 8th point
		// lower right on upper shelf
		pt[g].set(x, ty, -z);
		norm[g].set( pt[g].x, pt[g].y, pt[g].z );
		norm[g].normalize();
		face[f].vert[0].normIndex = g;
		face[f].vert[0].vertIndex = g;
		face[f].vert[4].normIndex = g;
		face[f].vert[4].vertIndex = g;
		g++;

		f++;

		/**************************/
		/****** SIDE FACES ********/
		/**************************/

		// g is 8 more than the first point
		
		// {1, 5, 8, 4, 1}
		face[f].nVerts = 5;
		face[f].vert[4].normIndex = g-8+0;
		face[f].vert[4].vertIndex = g-8+0;
		face[f].vert[3].normIndex = g-8+4;
		face[f].vert[3].vertIndex = g-8+4;
		face[f].vert[2].normIndex = g-8+7;
		face[f].vert[2].vertIndex = g-8+7;
		face[f].vert[1].normIndex = g-8+3;
		face[f].vert[1].vertIndex = g-8+3;
		face[f].vert[0].normIndex = g-8+0;
		face[f].vert[0].vertIndex = g-8+0;
		f++;

		// {2, 6, 5, 1, 2}
		face[f].nVerts = 5;
		face[f].vert[4].normIndex = g-8+1;
		face[f].vert[4].vertIndex = g-8+1;
		face[f].vert[3].normIndex = g-8+5;
		face[f].vert[3].vertIndex = g-8+5;
		face[f].vert[2].normIndex = g-8+4;
		face[f].vert[2].vertIndex = g-8+4;
		face[f].vert[1].normIndex = g-8+0;
		face[f].vert[1].vertIndex = g-8+0;
		face[f].vert[0].normIndex = g-8+1;
		face[f].vert[0].vertIndex = g-8+1;
		f++;

		// {3, 7, 6, 2, 3}
		face[f].nVerts = 5;
		face[f].vert[4].normIndex = g-8+2;
		face[f].vert[4].vertIndex = g-8+2;
		face[f].vert[3].normIndex = g-8+6;
		face[f].vert[3].vertIndex = g-8+6;
		face[f].vert[2].normIndex = g-8+5;
		face[f].vert[2].vertIndex = g-8+5;
		face[f].vert[1].normIndex = g-8+1;
		face[f].vert[1].vertIndex = g-8+1;
		face[f].vert[0].normIndex = g-8+2;
		face[f].vert[0].vertIndex = g-8+2;
		f++;

		// {4, 8, 7, 3, 4}
		face[f].nVerts = 5;
		face[f].vert[4].normIndex = g-8+3;
		face[f].vert[4].vertIndex = g-8+3;
		face[f].vert[3].normIndex = g-8+7;
		face[f].vert[3].vertIndex = g-8+7;
		face[f].vert[2].normIndex = g-8+6;
		face[f].vert[2].vertIndex = g-8+6;
		face[f].vert[1].normIndex = g-8+2;
		face[f].vert[1].vertIndex = g-8+2;
		face[f].vert[0].normIndex = g-8+3;
		face[f].vert[0].vertIndex = g-8+3;
		f++;
	}

	numFaces = f;
}

// EPN: added parameter
void Mesh::draw(int wire)
{
	for (int f = 0; f < numFaces; f++)
	{
		// EPN:  added simple test to flip between simple solid
		// to wore frame.
		if (wire == 0)
			glBegin(GL_POLYGON);
		else
			glBegin(GL_LINE_LOOP);
		for (int v = 0; v < face[f].nVerts; v++)
		{
			int in = face[f].vert[v].normIndex;
			int iv = face[f].vert[v].vertIndex;
			glNormal3f(norm[in].x, norm[in].y, norm[in].z);
			glVertex3f(pt[iv].x, pt[iv].y, pt[iv].z);
		}
		glEnd();
	}
}


// Create display list for improved performance
// http://www.lighthouse3d.com/opengl/displaylists/index.php3?3
GLuint Mesh::createDL(int wire)
{
	GLuint objectDL;

	// Create the id for the list
	objectDL = glGenLists(1);

	// start list
	glNewList(objectDL,GL_COMPILE);
	{
		// call the function that contains the rendering commands
		draw(wire);

	} // end list
	glEndList();

	return(objectDL);
}